from copy import deepcopy
from time import time
import numpy as np
import jax
import jax.numpy as jnp
import jax.random as jr
from jax.scipy.special import logsumexp
import numpy.typing as npt
from .utils import _solve_mbar

jax.config.update("jax_enable_x64", True)

__version__ = "1.4.5"

## A small diagonal matrix with __EPS__ as its diagonal elements is added
## to the Hessian matrix to avoid the case where the Hessian matrix is singular
## due to underflow.
__EPS__ = 1e-16


class FastMBAR:
    """
    The FastMBAR class is initialized with an energy matrix and an array
    of num of conformations. The corresponding MBAR equation is solved
    in the constructor. Therefore, the relative free energies of states
    used in the energy matrix is calculated in the constructor. The
    method **calculate_free_energies_for_perturbed_states**
    can be used to calculated the relative free energies of perturbed states.
    """

    def __init__(
        self,
        energy: npt.NDArray[np.float64],
        num_conf: npt.NDArray[np.int64],
        bootstrap: bool = False,
        bootstrap_block_size: int = 3,
        bootstrap_num_rep: int = 100,
        verbose: bool = False,
        method: str = "Newton",
    ) -> None:
        """Initializer for the class FastMBAR

        Parameters
        ----------
        energy : 2D ndarray
            It has a size of (M x N), where M is the number of states
            and N is the total number of conformations. The entry energy[i,j]
            is the reduced (unitless) energy of conformation j in state i.
            If bootstrapping is used to calculate the uncertainty, the order
            of conformations matters. Conformations sampled from one state
            need to occpy a continuous chunk of collumns. Conformations sampled
            from state k need to occupy collumns to the left of conformations
            sampled from state l if k < l. If bootstrapping is not used, then
            the order of conformation does not matter.
        num_conf: 1D int ndarray
            It should have a size of M, where num_conf[i] is the num of
            conformations sampled from state i. Therefore, np.sum(num_conf)
            has to be equal to N. All entries in num_conf have to be strictly
            greater than 0.
        bootstrap: bool, optional (default=False)
            If bootstrap is True, the uncertainty of the calculated free energies
            will be estimate using block bootstraping.
        bootstrap_block_size: int, optional (default=3)
            block size used in block bootstrapping
        bootstrap_num_rep: int, optional (default=100)
            number of repreats in block bootstrapping
        verbose: bool, optional (default=False)
            if verbose is true, the detailed information of solving MBAR equations
            is printed.
        method: str, optional (default="Newton")
            the method used to solve the MBAR equation. The default is Newton's method.
        """

        #### check the parameters: energy and num_conf
        ## Note that both self.energy and self.conf will be on CPUs no matter
        ## if cuda is True or False.

        ## check energy
        if isinstance(energy, np.ndarray):
            if energy.ndim != 2:
                raise ValueError("energy has to be two dimensional")
            self.energy = energy.astype(np.float64)
        else:
            raise TypeError("energy has to be a 2-D ndarray")

        ## check num_conf
        if isinstance(num_conf, np.ndarray):
            if num_conf.ndim != 1:
                raise ValueError("num_conf has to be one dimensional")
            self.num_conf = num_conf.astype(np.float64)
        else:
            raise TypeError("num_conf has to be a 1-D ndarray.")

        ## check the shape of energy and num_conf
        if energy.shape[0] != num_conf.shape[0]:
            raise ValueError(
                "The number of rows in energy must be equal to the length of num_conf. "
                f"Got {energy.shape[0]} rows in energy and {num_conf.shape[0]} elements in num_conf."
            )

        ## check if the number of conformations sampled from each state is greater than 0
        if np.any(self.num_conf <= 0):
            raise ValueError(
                "all entries in num_conf have to be strictly greater than 0."
            )

        ## check if the total number of conformations is equal to the number of columns in energy
        total_num_conf = np.sum(self.num_conf)
        num_columns = self.energy.shape[1]
        if total_num_conf != num_columns:
            raise ValueError(
                f"The sum of num_conf ({total_num_conf}) must be equal to the number of columns in energy ({num_columns})."
            )

        self.M = self.energy.shape[0]
        self.N = self.energy.shape[1]

        ## whether to use bootstrap to estimate the uncertainty of the calculated free energies
        ## bootstrap needs to be a boolean
        if not isinstance(bootstrap, bool):
            raise TypeError("bootstrap has to be a boolean.")
        self.bootstrap = bootstrap

        ## block size used in block bootstrapping
        if not isinstance(bootstrap_block_size, int):
            raise TypeError("bootstrap_block_size has to be an integer.")
        self.bootstrap_block_size = bootstrap_block_size

        ## number of repeats in block bootstrapping
        if not isinstance(bootstrap_num_rep, int):
            raise TypeError("bootstrap_num_rep has to be an integer.")
        self.bootstrap_num_rep = bootstrap_num_rep

        ## whether to print the detailed information of solving MBAR equations
        ## verbose needs to be a boolean
        if not isinstance(verbose, bool):
            raise TypeError("verbose has to be a boolean.")
        self.verbose = verbose

        ## method used to solve the MBAR equation
        if not isinstance(method, str):
            raise TypeError("method has to be a string.")
        if method not in ["Newton", "L-BFGS-B"]:
            raise ValueError("method has to be Newton or L-BFGS-B.")
        self.method = method

        # ## solve the MBAR equation
        if self.bootstrap is False:
            dF_init = jnp.zeros(self.M - 1)
            dF = _solve_mbar(
                dF_init,
                self.energy,
                self.num_conf,
                self.method,
                verbose=self.verbose,
            )

            ## shift self._F such that \sum_i F[i]*num_conf[i] = 0
            F = jnp.concat([jnp.zeros(1), dF])
            pi = self.num_conf / self.N
            self._F = F - jnp.sum(pi * F)

            b = -self._F - jnp.log(self.num_conf)
            self._log_prob_mix = logsumexp(-(self.energy + b[:, None]), axis=0)
            self._log_mu = -self.log_prob_mix

            ## Compute self._F_cov under the constraint that \sum_i F[i]*num_conf[i] = 0

            ## There are two ways to compute the covariance matrix of self._F
            ## See equation 4.2 and 6.4 in the following paper for details
            ## "Kong, A., et al. "A theory of statistical models for Monte Carlo integration."
            ## Journal of the Royal Statistical Society Series B: Statistical Methodology 65.3
            ## (2003): 585-604." https://doi.org/10.1111/1467-9868.00404

            ## The first way as shown in the following uses equation 4.2
            ## this method is more general than the following method, meaning that
            ## it can also be used to compute covariance matrix for perturbed states.
            ## Therefore it is used here.

            self.P = jnp.exp(
                -(self.energy - self._F[:, None] + self._log_prob_mix)
            ).transpose()
            W = jnp.diag(self.num_conf)
            Q, R = jnp.linalg.qr(self.P)
            A = jnp.eye(self.M, device=W.device) - R @ W @ R.T
            self._F_cov = R.T @ jnp.linalg.pinv(A, hermitian=True, rtol=1e-7) @ R

            # # The second way uses equation 6.4
            # if self._batch_size is None:
            #     H = _compute_hessian_log_likelihood_of_F(
            #         self._F, self.energy, self.num_conf
            #     )
            # else:
            #     H = _compute_hessian_log_likelihood_of_F_in_batch(
            #         self._F, self.energy, self.num_conf, self._batch_size
            #     )
            # Hp = H.new_zeros((self.M + 1, self.M + 1)).cpu()
            # Hp[0 : self.M, 0 : self.M] = H
            # Hp[self.M, 0 : self.M] = -self.num_conf
            # Hp[0 : self.M, self.M] = -self.num_conf
            # Hp[self.M, self.M] = 0
            # self._F_cov = torch.linalg.inv(-Hp)[0 : self.M, 0 : self.M]
            # self._F_cov = self._F_cov - torch.diag(1 / self.num_conf) + 1 / self.N

            self._F_std = jnp.sqrt(self._F_cov.diagonal())

            self._DeltaF = self._F[None, :] - self._F[:, None]
            self._DeltaF_cov = (
                self._F_cov.diagonal()[:, None]
                + self._F_cov.diagonal()[None, :]
                - 2 * self._F_cov
            )
            self._DeltaF_std = jnp.sqrt(self._DeltaF_cov)

        elif self.bootstrap is True:
            dFs = []
            log_prob_mix = []
            dF_init = jnp.zeros(self.M - 1)
            self._conf_idx = []
            for _ in range(self.bootstrap_num_rep):
                conf_idx = _bootstrap_conf_idx(num_conf, self.bootstrap_block_size)
                dF = _solve_mbar(
                    dF_init,
                    self.energy[:, conf_idx],
                    self.num_conf,
                    self.method,
                    verbose=self.verbose,
                )
                dF_init = dF
                dFs.append(deepcopy(dF))

                F = jnp.concat([jnp.zeros(1), dF])
                ## shift F such that \sum_i F[i]*num_conf[i] = 0
                pi = self.num_conf / self.N
                F = F - jnp.sum(pi * F)

                b = -F - jnp.log(self.num_conf / self.N)
                log_prob_mix.append(
                    logsumexp(-(self.energy[:, conf_idx] + b[:, None]), axis=0)
                )
                self._conf_idx.append(conf_idx)

            dF = jnp.stack(dFs, axis=1)
            F = jnp.concat([jnp.zeros((1, dF.shape[1])), dF], axis=0)

            ## shift F such that \sum_i F[i]*num_conf[i] = 0
            pi = self.num_conf / self.N
            self._F_bootstrap = F - jnp.sum(pi[:, None] * F, axis=0)

            self._F = jnp.mean(self._F_bootstrap, axis=1)
            self._F_std = jnp.std(self._F_bootstrap, axis=1)
            self._F_cov = jnp.cov(self._F_bootstrap)

            self._log_prob_mix = jnp.stack(log_prob_mix, axis=0)
            DeltaF = self._F_bootstrap[None, :, :] - self._F_bootstrap[:, None, :]
            self._DeltaF = jnp.mean(DeltaF, axis=2)
            self._DeltaF_std = jnp.std(DeltaF, axis=2)

    @property
    def F(self) -> np.ndarray:
        r"""Free energies of the states under the constraint :math:`\sum_{k=1}^{M} N_k * F_k = 0`,
        where :math:`N_k` is the number of conformations sampled from state k.
        """
        return np.array(self._F)

    @property
    def F_std(self) -> np.ndarray:
        r"""Standard deviation of the free energies of the states under the constraint
        :math:`\sum_{k=1}^{M} N_k * F_k = 0`,
        where :math:`N_k` is the number of conformations sampled from state k.
        """
        return np.array(self._F_std)

    @property
    def F_cov(self) -> np.ndarray:
        r"""Covariance matrix of the free energies of the states under the constraint
        :math:`\sum_{k=1}^{M} N_k * F_k = 0`,
        where :math:`N_k` is the number of conformations sampled from state k.
        """
        return np.array(self._F_cov)

    @property
    def DeltaF(self) -> np.ndarray:
        r"""Free energy difference between states.
        :math:`\mathrm{DeltaF}[i,j]` is the free energy difference between state j and state i,
        i.e., :math:`\mathrm{DeltaF}[i,j] = F[j] - F[i]` .
        """
        return np.array(self._DeltaF)

    @property
    def DeltaF_std(self) -> np.ndarray:
        r"""Standard deviation of the free energy difference between states.
        :math:`\mathrm{DeltaF_std}[i,j]` is the standard deviation of the free energy
        difference :math:`\mathrm{DeltaF}[i,j]`.
        """
        return np.array(self._DeltaF_std)

    @property
    def log_prob_mix(self) -> np.ndarray:
        """the log probability density of conformations in the mixture distribution."""
        return np.array(self._log_prob_mix)

    def calculate_free_energies_of_perturbed_states(
        self, energy_perturbed: npt.NDArray[np.float64]
    ) -> dict:
        r"""calculate free energies for perturbed states.

        Parameters
        -----------
        energy_perturbed: 2-D float ndarray with size of (L,N)
            each row of the energy_perturbed matrix represents a state and
            the value energy_perturbed[l,n] represents the reduced energy
            of the n'th conformation in the l'th perturbed state.

        Returns
        -------
        results: dict
            a dictionary containing the following keys:

            **F** - the free energy of the perturbed states.

            **F_std** - the standard deviation of the free energy of the perturbed states.

            **F_cov** - the covariance between the free energies of the perturbed states.

            **DeltaF** - :math:`\mathrm{DeltaF}[k,l]` is the free energy difference between state
            :math:`k` and state :math:`l`, i.e., :math:`\mathrm{DeltaF}[k,l] = F[l] - F[k]` .

            **DeltaF_std** - the standard deviation of the free energy difference.
        """

        if isinstance(energy_perturbed, np.ndarray):
            energy_perturbed = energy_perturbed.astype(np.float64)
        else:
            raise TypeError("energy_perturbed has to be a numpy array")

        if energy_perturbed.ndim != 2:
            raise ValueError(
                "energy_perturbed has to be a 2-D ndarray or a 2-D tensor."
            )
        if energy_perturbed.shape[1] != self.energy.shape[1]:
            raise ValueError(
                "the number of columns in energy_perturbed has to be equal to the number of columns in energy."
            )

        L = energy_perturbed.shape[0]

        F = None
        F_cov = None
        F_std = None

        if not self.bootstrap:
            du = energy_perturbed + self._log_prob_mix
            F = -logsumexp(-du, axis=1)

            F_ext = jnp.concat([self._F, F])
            U = jnp.concat([self.energy, energy_perturbed], axis=0).transpose()
            self._P = jnp.exp(-(U - F_ext + self._log_prob_mix[:, None]))

            W = jnp.diag(jnp.concat([self.num_conf, jnp.zeros(L)]))
            Q, R = jnp.linalg.qr(self._P)
            A = jnp.eye(self.M + L, device=W.device) - R @ W @ R.T
            F_cov = R.T @ jnp.linalg.pinv(A, hermitian=True, rtol=1e-8) @ R

            F_cov = F_cov[self.M :, self.M :]
            F_std = jnp.sqrt(F_cov.diagonal())

            DeltaF = F[None, :] - F[:, None]
            DeltaF_cov = (
                F_cov.diagonal()[:, None] + F_cov.diagonal()[None, :] - 2 * F_cov
            )
            DeltaF_std = jnp.sqrt(DeltaF_cov)

        else:
            F_list = []
            for k in range(self.bootstrap_num_rep):
                du = energy_perturbed[:, self._conf_idx[k]] + self._log_prob_mix[k]
                F = -logsumexp(-du, axis=1)
                F_list.append(F)
            Fs = jnp.stack(F_list, axis=1)
            F = jnp.mean(Fs, axis=1)
            F_std = jnp.std(Fs, axis=1)
            F_cov = jnp.cov(Fs)

            DeltaF = F[None, :] - F[:, None]
            DeltaF_std = jnp.std(Fs[:, :, None] - Fs[:, None, :], axis=1)

        results = {
            "F": np.array(F),
            "F_std": np.array(F_std),
            "F_cov": np.array(F_cov),
            "DeltaF": np.array(DeltaF),
            "DeltaF_std": np.array(DeltaF_std),
        }

        return results





def _bootstrap_conf_idx(num_conf, bootstrap_block_size):
    num_conf_cumsum = jnp.cumsum(num_conf, axis=0).tolist()
    num_conf_cumsum.pop(-1)
    num_conf_cumsum = [0] + num_conf_cumsum
    conf_idx = []

    key = jr.PRNGKey(int(time() * 1000))
    for i in range(len(num_conf)):
        len_seq = int(num_conf[i])  ## using circular block bootstrap
        num_sample_block = int(jnp.ceil(len_seq / bootstrap_block_size))
        key, subkey = jr.split(key)
        idxs = jr.randint(subkey, (num_sample_block,), 0, len_seq)
        sample_idx = jnp.concat(
            [jnp.arange(idx, idx + bootstrap_block_size) for idx in idxs]
        )
        sample_idx = jnp.remainder(sample_idx, len_seq)
        sample_idx = sample_idx[0:len_seq]
        sample_idx = sample_idx + num_conf_cumsum[i]
        conf_idx.append(sample_idx)
    conf_idx = jnp.concat(conf_idx)
    return conf_idx



from .bayesmbar import BayesMBAR
from .cbayesmbar import CBayesMBAR
from .bayesbar import BayesBAR
from .fastmbar import FastMBAR

__all__ = ('FastMBAR', 'BayesMBAR', 'BayesBAR', 'CBayesMBAR', '__init__')   



from collections.abc import Sequence
from typing import List, Tuple
import time
import numpy as np
from numpy import ndarray
import networkx as nx
import jax
import jax.numpy as jnp
from jax import random
from jax import hessian, jit, value_and_grad, vmap
from scipy import optimize
from .bayesmbar import _sample_from_logdensity
from .utils import fmin_newton, _compute_log_likelihood_of_dF

jax.config.update("jax_enable_x64", True)


class CBayesMBAR:
    """
    Coupled BayesMBAR
    """

    def __init__(
        self,
        energies: List[ndarray],
        nums_conf: List[ndarray],
        identical_states: List[List[Tuple[int, int]]],
        sample_size: int = 1000,
        warmup_steps: int = 500,
        method: str = "Newton",
        random_seed: int = None,
        verbose: bool = True,
    ):
        """
        Parameters
        ----------
        energies : List[ndarray]
            A list of energies for all coupled MBAR systems. The energies should be in the unit of kT.
        nums_conf : List[ndarray]
            A list of the number of configurations for all coupled MBAR systems.
        identical_states : List[List[(int, int)]]
            A list of identical states. Each element in the outer list is a list of tuples and
            represents a group of states that are identical to each other. States are represented
            by a tuple where the first element is the index of a MBAR system and the second element
            is the index of the state in that MBAR system. For example,
            identical_states = [[(0, 3), (1, 0)], [(1, 3), (2, 0), (3, 0)]] means that state 3 in
            system 0 is identical to state 0 in system 1, and state 3 in system 1 is identical to
            state 0 in system 2, and state 0 in system 3.
        sample_size : int, optional
            The number of samples to draw from the likelihood. The default is 1000.
        warmup_steps : int, optional
            The number of warmup steps for the HMC sampler. The default is 500.
        method : str, optional
            The optimization method for finding the mode of the likelihood. Options are "Newton" or "L-BFGS-B". The default is "Newton".
        random_seed : int, optional
            The random seed. The default is None, which means the random seed is generated from the current time.
        verbose : bool, optional
            Whether to print out the progress of the sampling. The default is True.
        """
        self._energies = [jnp.float64(u) for u in energies]
        self._nums_conf = [jnp.int32(n) for n in nums_conf]
        self._identical_states = identical_states

        self._sample_size = sample_size
        self._warmup_steps = warmup_steps

        self._verbose = verbose
        if random_seed is None:
            random_seed = int(time.time())
        self._rng_key = jax.random.PRNGKey(int(time.time()))

        # number of states in each mbar system
        self._nums_state = [len(s) for s in self._nums_conf]

        self._Q = _compute_projection(self._nums_state, identical_states)
        x = jnp.zeros((self._Q.shape[1]))

        print("Solve for the mode of the likelihood")
        if method == "Newton":
            f = jit(value_and_grad(_compute_cmbar_loss_likelihood))
            hess = jit(hessian(_compute_cmbar_loss_likelihood))
            res = fmin_newton(
                f,
                hess,
                x,
                args=(self._Q, self._energies, self._nums_conf),
            )
        elif method == "L-BFGS-B":
            options = {"disp": verbose, "gtol": 1e-8}
            f = jit(value_and_grad(_compute_cmbar_loss_likelihood))
            res = optimize.minimize(
                lambda x: [
                    np.array(r) for r in f(x, self._Q, self._energies, self._nums_conf)
                ],
                x,
                jac=True,
                method="L-BFGS-B",
                tol=1e-12,
                options=options,
            )
        else:
            raise ValueError("Invalid method")

        x_mode_ll = res["x"]
        self._dF_mode_ll = jnp.dot(self._Q, x_mode_ll)
        self._state_F_mode_ll = _dF_to_state_F(self._dF_mode_ll, self._nums_state)

        if self._sample_size > 0:
            print("=====================================================")
            print("Sample from the likelihood")

            self._rng_key, subkey = random.split(self._rng_key)

            def logdensity(x):
                return _compute_cmbar_log_likelihood(
                    x, self._Q, self._energies, self._nums_conf
                )

            self._x_samples_ll = _sample_from_logdensity(
                subkey,
                x_mode_ll,
                logdensity,
                self._warmup_steps,
                self._sample_size,
                self._verbose,
            )
            self._dF_samples_ll = self._x_samples_ll @ self._Q.T

            self._state_F_samples_ll = vmap(_dF_to_state_F, in_axes=[0, None])(
                self._dF_samples_ll, self._nums_state
            )
            self._state_F_mean_ll = [
                jnp.mean(F, axis=0) for F in self._state_F_samples_ll
            ]

    @property
    def F_mode(self) -> List[ndarray]:
        """ The mode of free energies of all states in all MBAR systems. The free energy of 
        state 0 in each system is set to 0.
        """
        return [
            np.array(jax.device_put(F, jax.devices("cpu")[0]))
            for F in self._state_F_mode_ll
        ]

    @property
    def F_samples(self) -> List[ndarray]:
        """ The samples of free energies of all states in all MBAR systems. The free energy of
        state 0 in each system is set to 0.
        """
        return [
            np.array(jax.device_put(F, jax.devices("cpu")[0]))
            for F in self._state_F_samples_ll
        ]

    @property
    def F_mean(self) -> List[ndarray]:
        """ The mean of free energies of all states in all MBAR systems. The free energy of
        state 0 in each system is set to 0.
        """
        return [
            np.array(jax.device_put(F, jax.devices("cpu")[0]))
            for F in self._state_F_mean_ll
        ]

    @property
    def DeltaF_mode(self) -> List[ndarray]:
        """ The mode of free energy differences between all pairs of states in every MBAR system.
        """
        return [F[None, :] - F[:, None] for F in self.F_mode]

    @property
    def DeltaF_mean(self) -> List[ndarray]:
        """ The mean of free energy differences between all pairs of states in every MBAR system.
        """
        return [F[None, :] - F[:, None] for F in self.F_mean]

    @property
    def DeltaF_std(self) -> List[ndarray]:
        """ The standard deviation of free energy differences between all pairs of states in every MBAR system.
        """
        return [np.std(F[:, None, :] - F[:, :, None], 0) for F in self.F_samples]


def _dF_to_state_F(dF: jnp.ndarray, nums_state: List[int]) -> List[jnp.ndarray]:
    state_dF = []
    idx = 0
    for n in nums_state:
        state_dF.append(dF[idx : idx + n - 1])
        idx += n - 1
    state_F = [jnp.concatenate([jnp.zeros(1), dF]) for dF in state_dF]
    return state_F


def _F_to_state_F(
    F: jnp.ndarray, nums_state: Sequence[int], state_idx_to_F_idx: dict[(int, int), int]
) -> list[jnp.ndarray]:
    F_of_states = []
    for i in range(len(nums_state)):
        idx = jnp.array([state_idx_to_F_idx[(i, j)] for j in range(nums_state[i])])
        F_of_states.append(F[..., idx])
    return F_of_states


def _compute_cmbar_log_likelihood(
    x: jnp.ndarray,
    Q: jnp.ndarray,
    energies: Sequence[jnp.ndarray],
    nums_conf: Sequence[jnp.ndarray],
):
    dF = jnp.dot(Q, x)
    nums_state = [len(s) for s in nums_conf]
    state_dF = []
    idx = 0
    for n in nums_state:
        state_dF.append(dF[..., idx : idx + n - 1])
        idx += n - 1

    log_likelihood = 0
    for energy, num_conf, dF in zip(energies, nums_conf, state_dF):
        log_likelihood += _compute_log_likelihood_of_dF(dF, energy, num_conf)

    return log_likelihood


def _compute_cmbar_loss_likelihood(
    x: jnp.ndarray,
    Q: jnp.ndarray,
    energies: Sequence[jnp.ndarray],
    nums_conf: Sequence[jnp.ndarray],
):
    log_likelihood = _compute_cmbar_log_likelihood(x, Q, energies, nums_conf)
    N = jnp.sum(jnp.array(nums_conf))
    return -log_likelihood / N


def _generate_dF_graph(
    nums_state: List[int], identical_states: List[List[Tuple[int, int]]]
) -> nx.Graph:
    """
    Generate a graph where each node represents a state and each edge represents a perturbation
    between two states whose free energy difference is to be computed. For states that are identical
    as specified in identical_states, edges are added between all pairs of states that are identical.
    We add edges between all pairs of identical states because we will use them to constraint
    the free energy difference of these edges to be zero.

    Parameters
    ----------
    nums_state : Sequence[int]
        A list of the number of states in each MBAR system.
    identical_states : Sequence[Sequence[(int, int)]]
        Same as the identical_states parameter in the CBayesMBAR class.
    """
    G = nx.Graph()
    for i in range(len(nums_state)):
        for j in range(1, nums_state[i]):
            G.add_edge((i, 0), (i, j))

    for states in identical_states:
        for i in range(len(states)):
            for j in range(i + 1, len(states)):
                G.add_edge(states[i], states[j])

    return G


def _compute_projection(
    nums_state: List[int], identical_states: List[List[Tuple[int, int]]]
) -> ndarray:
    """Compute the projection matrix Q

    This function converts contraints on the free energy differences due to identical states
    to a projection matrix Q that does the following:

    dF = Q @ x, where x is a vector of independent variables without any constraints.
    dF is the vector of all free energy differences between states in all MBAR systems that satisfy
    the constraints because it is generated by the projection matrix Q.

    """

    ## map from state index (i,j), where i is the index of the MBAR system and j is the index of
    ## the state in that system, to the index of the free energy differences in the vector dF
    ## i.e., dF[state_idx_to_dF_idx[(i,j)]] is the free energy difference between state j and
    ## state 0 in system i
    state_dF_idx_to_dF_idx = {}
    idx = 0
    for i in range(len(nums_state)):
        for j in range(1, nums_state[i]):
            state_dF_idx_to_dF_idx[(i, j)] = idx
            idx += 1

    ## build the dF graph and find the cycles in the graph
    graph = _generate_dF_graph(nums_state, identical_states)
    cycles = nx.cycle_basis(graph)

    ## build the matrix A where each row corresponds to a cycle in the graph
    ## For each cycle, we build a vector q that has the same length as the number of free energy
    ## differences.
    ## q[k] = 1 if the k-th free energy difference is in the cycle and the second state
    ## in the edge is the reference state (state 0).
    ## q[k] = -1 if the k-th free energy difference is in the cycle and the first state
    ## in the edge is the reference state.
    ## q[k] = 0 if the k-th free energy difference is not in the cycle.
    ## Note that because all free energy differences are within the same MBAR system,
    ## if a cycle contains an edge between two states in different systems, that edge
    ## must have been added because the two states are identical. In this case, we set q[k] = 0.

    A = []
    for cycle in cycles:
        q = np.zeros(len(state_dF_idx_to_dF_idx))
        for i in range(len(cycle)):
            n1, n2 = cycle[i], cycle[(i + 1) % len(cycle)]

            ## nodes are represented by tuples (m, k), where m is the index of the MBAR system
            ## and k is the index of the state in that system

            ## if the two nodes are in different systems, they must be identical states
            ## and we set q[k] = 0
            if n1[0] != n2[0]:
                continue

            ## check that one of the states is the reference state (state 0)
            assert n1[1] == 0 or n2[1] == 0

            ## set q[k] = 1 if the second state in the edge is the reference state
            ## set q[k] = -1 if the first state in the edge is the reference state
            if n1[1] == 0 and n2[1] != 0:
                q[state_dF_idx_to_dF_idx[n2]] = 1
            elif n1[1] != 0 and n2[1] == 0:
                q[state_dF_idx_to_dF_idx[n1]] = -1
            else:
                raise ValueError("Invalid cycle")

        if np.any(q != 0):
            A.append(q)

    ## compute the projection matrix Q using the QR decomposition
    B = np.array(A).T
    Q, R = np.linalg.qr(B, mode="complete")
    Q = Q[:, B.shape[1] :]

    return Q


def _map_from_state_idx_to_F_idx(
    nums_state: Sequence[int], identical_states: Sequence[Sequence[(int, int)]]
) -> dict[(int, int), int]:
    ## convert the list of identical states to a dictionary where the key is the state index
    ## and the value is the list of states that are identical to the key state
    iden_states_dict = {}
    for states in identical_states:
        for i in range(len(states)):
            iden_states_dict[states[i]] = [
                states[j] for j in range(len(states)) if j != i
            ]

    ## loop over all states and map each state to an index in the free energy F
    F_idx = 0
    state_idx_to_F_idx = {}
    for i in range(len(nums_state)):
        for j in range(nums_state[i]):
            state_idx = (i, j)

            ## if not other states are identical to this state, assign a new index
            if state_idx not in iden_states_dict.keys():
                state_idx_to_F_idx[state_idx] = F_idx
                F_idx += 1

            ## if there are other states that are identical to this state, go over the list of identical states and check if any of them has been assigned an index. If so, assign the same index to this state. Otherwise, assign a new index
            else:
                for s in iden_states_dict[state_idx]:
                    if s in state_idx_to_F_idx.keys():
                        state_idx_to_F_idx[state_idx] = state_idx_to_F_idx[s]
                        break
                if state_idx not in state_idx_to_F_idx.keys():
                    state_idx_to_F_idx[state_idx] = F_idx
                    F_idx += 1

    return state_idx_to_F_idx



from functools import partial
from typing import Literal

import numpy as np
from numpy.typing import NDArray
import jax

import jax.numpy as jnp
from jax import hessian, jit, value_and_grad
from jax import random
import blackjax
from optax import sgd
import optax
from .utils import (
    _solve_mbar,
    fmin_newton,
    _compute_log_likelihood_of_dF,
    _compute_log_likelihood_of_F,
    _compute_loss_likelihood_of_dF,
)

jax.config.update("jax_enable_x64", True)


class BayesMBAR:
    """Bayesian Multistate Bennett Acceptance Ratio (BayesMBAR) method"""

    def __init__(
        self,
        energy: np.ndarray,
        num_conf: np.ndarray,
        prior: Literal["uniform", "normal"] = "uniform",
        mean: Literal["constant", "linear", "quadratic"] = "constant",
        kernel: Literal["SE", "Matern52", "Matern32", "RQ"] = "SE",
        state_cv: np.ndarray = None,
        sample_size: int = 1000,
        warmup_steps: int = 500,
        optimize_steps: int = 10000,
        verbose: bool = True,
        random_seed: int = 0,
        method: Literal["Newton", "L-BFGS-B"] = "Newton",
    ) -> None:
        """
        Args:
            energy (np.ndarray): Energy matrix of shape (m, n), where m is the number of states and n is the number of configurations.
            num_conf (np.ndarray): Number of configurations in each state. It is a 1D array of length m.
            prior (str, optional): Prior distribution of dF. It can be either "uniform" or "normal". Defaults to "uniform".
            mean (str, optional): Mean function of the prior. It can be either "constant", "linear", or "quadratic". Defaults to "constant".
            kernel (str, optional): Kernel function of the prior. It can be either "SE", "Matern52", "Matern32", or "RQ". Defaults to "SE".
            state_cv (np.ndarray, optional): State collective variables. It is a 2D array of shape (n, d), where n is the number of configurations and d is the dimension of the collective variables. Defaults to None.
            sample_size (int, optional): Number of samples drawn from the posterior distribution. Defaults to 1000.
            warmup_steps (int, optional): Number of warmup steps used to find the step size and mass matrix of the NUTS sampler. Defaults to 500.
            optimize_steps (int, optional): Number of optimization steps used to learn the hyperparameters when normal priors are used. Defaults to 10000.
            verbose (bool, optional): Whether to print the progress bar for the optimization and sampling. Defaults to True.
            random_seed (int, optional): Random seed. Defaults to 0.

        """

        self._energy = jnp.float64(energy)
        self._num_conf = jnp.int32(num_conf)

        self._prior = prior
        self._mean_name = mean
        self._kernel_name = kernel

        if state_cv is not None:
            self._state_cv = state_cv[1:]

        self._sample_size = sample_size
        self._warmup_steps = warmup_steps
        self._optimize_steps = optimize_steps

        self._verbose = verbose
        self._method = method
        self._rng_key = jax.random.PRNGKey(random_seed)

        self._m = self._energy.shape[0]
        self._n = self._energy.shape[1]

        # We first compute the mode estimate based on the likelihood
        # because it is used in both the uniform and normal priors.
        # The mode estimate based on the likelihood is the solution to the MBAR equation.

        print("Solve for the mode of the likelihood")
        dF_init = jnp.zeros((self._m - 1,))
        dF = _solve_mbar(
            dF_init, self._energy, self._num_conf, self._method, self._verbose
        )

        # f = jit(value_and_grad(_compute_loss_likelihood_of_dF))
        # hess = jit(hessian(_compute_loss_likelihood_of_dF))
        # res = fmin_newton(f, hess, dF_init, args=(self._energy, self._num_conf))
        # dF = res["x"]

        self._dF_mode_ll = dF

        # sample dF based on the likelihood.
        # When the uniform prior is used, the posterior distribution of dF is
        # the same as the likelihood function.
        # Thefore so these samples are also samples from the posterior
        # distribution of dF when the uniform prior is used.

        print("=====================================================")
        print("Sample from the likelihood")

        self._rng_key, subkey = random.split(self._rng_key)

        def logdensity(dF):
            return _compute_log_likelihood_of_dF(dF, self._energy, self._num_conf)

        self._dF_samples_ll = _sample_from_logdensity(
            subkey,
            self._dF_mode_ll,
            logdensity,
            self._warmup_steps,
            self._sample_size,
            self._verbose,
        )

        ## compute the mean, covariance, and precision of dF based on the samples from the likelihood
        self._dF_mean_ll = jnp.mean(self._dF_samples_ll, axis=0)
        self._dF_cov_ll = jnp.cov(self._dF_samples_ll.T)

        L = jnp.linalg.cholesky(self._dF_cov_ll)
        L_inv = jax.scipy.linalg.solve_triangular(L, jnp.eye(L.shape[0]), lower=True)
        self._dF_prec_ll = L_inv.T.dot(L_inv)
        # self._dF_prec_ll = jnp.linalg.inv(self._dF_cov_ll)

        self._F_mode_ll = _dF_to_F(self._dF_mode_ll, self._num_conf)
        self._F_samples_ll = _dF_to_F(self._dF_samples_ll, self._num_conf)
        self._F_mean_ll = jnp.mean(self._F_samples_ll, axis=0)
        self._F_cov_ll = jnp.cov(self._F_samples_ll.T)

        ## we are done here if the prior is uniform.
        ## When normal prior is used, we need to learn the hyperparameters of the prior and then sample dF from the posterior distribution of dF.

        if self._prior == "normal":
            _data = {
                "energy": self._energy,
                "num_conf": self._num_conf,
                "dF_mean_ll": self._dF_mean_ll,
                "dF_prec_ll": self._dF_prec_ll,
                "state_cv": self._state_cv,
            }

            ## mean function of the prior
            if self._mean_name == "constant":
                self.mean_order = 0
                self.mean = partial(_mean, order=self.mean_order)
            elif self._mean_name == "linear":
                self.mean_order = 1
                self.mean = partial(_mean, order=self.mean_order)
            elif self._mean_name == "quadratic":
                self.mean_order = 2
                self.mean = partial(_mean, order=self.mean_order)

            ## learn the hyperparameters of the prior
            if self._kernel_name == "SE":
                self.kernel = _kernel_SE
            elif self._kernel_name == "Matern52":
                self.kernel = _kernel_Matern52
            elif self._kernel_name == "Matern32":
                self.kernel = _kernel_Matern32
            elif self._kernel_name == "RQ":
                self.kernel = _kernel_RQ

            ## initialize the hyperparameters based on the mode of the likelihood
            params = _init_params(
                self.mean_order,
                self._kernel_name,
                self._dF_mode_ll,
                self._state_cv,
                self._num_conf,
            )
            raw_params = _params_to_raw(params)

            ## optimize the hyperparameters
            self._rng_key, subkey = random.split(self._rng_key)
            optimizer = sgd(learning_rate=1e-3, momentum=0.9, nesterov=True)
            opt_state = optimizer.init(raw_params)

            @partial(jit, static_argnames=["mean", "kernel"])
            def step(key, raw_params, opt_state, mean, kernel, data):
                loss, grads = _compute_elbo_loss(key, raw_params, mean, kernel, data)
                update, opt_state = optimizer.update(grads, opt_state)
                raw_params = optax.apply_updates(raw_params, update)
                return loss, raw_params, opt_state

            for i in range(optimize_steps):
                loss, raw_params, opt_state = step(
                    subkey, raw_params, opt_state, self.mean, self.kernel, _data
                )
                self._rng_key, subkey = random.split(self._rng_key)
                if i % 100 == 0:
                    params = _params_from_raw(raw_params)
                    print(f"step: {i:>10d}, loss: {loss:10.4f}", _print_params(params))

            self._params = _params_from_raw(raw_params)
            self._dF_mean_prior = self.mean(self._params["mean"], self._state_cv)
            self._dF_cov_prior = self.kernel(
                self._params["kernel"],
                self._state_cv,
            )
            self._dF_prec_prior = jnp.linalg.inv(self._dF_cov_prior)

            ## solve for the mode of the posterior
            f = jit(value_and_grad(_compute_loss_joint_likelihood_of_dF))
            hess = jit(hessian(_compute_loss_joint_likelihood_of_dF))
            res = fmin_newton(
                f,
                hess,
                self._dF_mode_ll,
                args=(
                    self._energy,
                    self._num_conf,
                    self._dF_mean_prior,
                    self._dF_prec_prior,
                ),
            )
            self._dF_mode_posterior = res["x"]

            ## sample dF from the posterior
            def logdensity(dF):
                return _compute_log_joint_likelihood_of_dF(
                    dF,
                    self._energy,
                    self._num_conf,
                    self._dF_mean_prior,
                    self._dF_prec_prior,
                )

            self._rng_key, subkey = random.split(self._rng_key)
            self._dF_samples_posterior = _sample_from_logdensity(
                subkey,
                self._dF_mode_posterior,
                logdensity,
                self._warmup_steps,
                self._sample_size,
                self._verbose,
            )
            self._dF_mean_posterior = jnp.mean(self._dF_samples_posterior, axis=0)
            self._dF_cov_posterior = jnp.cov(self._dF_samples_posterior.T)
            self._dF_prec_posterior = jnp.linalg.inv(self._dF_cov_posterior)

            self._F_mode_posterior = _dF_to_F(self._dF_mode_posterior, self._num_conf)
            self._F_samples_posterior = _dF_to_F(
                self._dF_samples_posterior, self._num_conf
            )
            self._F_mean_posterior = jnp.mean(self._F_samples_posterior, axis=0)
            self._F_cov_posterior = jnp.cov(self._F_samples_posterior.T)

    @property
    def F_mode(self) -> NDArray:
        r"""The posterior mode estimate of the free energies of the states under the constraints that :math:`\sum_{k=1}^{M} N_k * F_k = 0`, where :math:`N_k` and :math:`F_k` are the number of conformations and the free energy of the k-th state, respectively."""
        if self._prior == "uniform":
            F_mode = self._F_mode_ll
        elif self._prior == "normal":
            F_mode = self._F_mode_posterior
        return np.array(jax.device_put(F_mode, jax.devices("cpu")[0]))

    @property
    def F_mean(self) -> NDArray:
        r"""The posterior mean of the free energies of the states under the constraints that :math:`\\sum_{k=1}^{M} N_k * F_k = 0`, where :math:`N_k` and :math:`F_k` are the number of conformations and the free energy of the k-th state, respectively."""

        if self._prior == "uniform":
            F_mean = self._F_mean_ll
        elif self._prior == "normal":
            F_mean = self._F_mean_posterior
        return np.array(jax.device_put(F_mean, jax.devices("cpu")[0]))

    @property
    def F_cov(self) -> NDArray:
        r"""The posterior covariance matrix of the free energies of the states under the constraints that :math:`\\sum_{k=1}^{M} N_k * F_k = 0`, where :math:`N_k` and :math:`F_k` are the number of conformations and the free energy of the k-th state, respectively."""
        if self._prior == "uniform":
            F_cov = self._F_cov_ll
        elif self._prior == "normal":
            F_cov = self._F_cov_posterior
        F_cov = F_cov - jnp.diag(1.0 / self._num_conf) + 1.0 / self._num_conf.sum()

        ## if the diagnoal elements of F_cov are negetive, set them to 1e-4
        condition = jnp.eye(F_cov.shape[0], dtype=bool) & (F_cov <= 0)
        F_cov = jnp.where(condition, 1e-4, F_cov)
        return np.array(jax.device_put(F_cov, jax.devices("cpu")[0]))

    @property
    def F_std(self) -> NDArray:
        r"""The posterior standard deviation of the free energies of the states under the constraints that :math:`\\sum_{k=1}^{M} N_k * F_k = 0`, where :math:`N_k` and :math:`F_k` are the number of conformations and the free energy of the k-th state, respectively."""
        return np.array(jnp.sqrt(jnp.diag(self.F_cov)))

    @property
    def F_samples(self) -> NDArray:
        """The samples of the free energies of the states from the posterior distribution under the constraints that :math:`\\sum_{k=1}^{M} N_k * F_k = 0`, where :math:`N_k` and :math:`F_k` are the number of conformations and the free energy of the k-th state, respectively."""
        if self._prior == "uniform":
            F_samples = self._F_samples_ll
        elif self._prior == "normal":
            F_samples = self._F_samples_posterior
        return np.array(jax.device_put(F_samples, jax.devices("cpu")[0]))

    @property
    def DeltaF_mode(self) -> NDArray:
        r"""The posterior mode estimate of free energy difference between states.
        DeltaF_mode[i,j] is the free energy difference between state :math:`j` and state :math:`i`,
        i.e., DeltaF_mode[i,j] = F_mode[j] - F_mode[i].
        """
        return self.F_mode[None, :] - self.F_mode[:, None]

    @property
    def DeltaF_mean(self) -> NDArray:
        r"""The posterior mean of free energy difference between states.
        DeltaF_mean[i,j] is the free energy difference between state :math:`j` and state :math:`i`,
        i.e., DeltaF_mean[i,j] = F_mean[j] - F_mean[i].
        """
        return self.F_mean[None, :] - self.F_mean[:, None]

    @property
    def DeltaF_std(self) -> NDArray:
        r"""The posterior standard deviation of free energy difference between states.
        DeltaF_std[i,j] is the posterior standard deviation of the free energy difference between state :math:`j` and state :math:`i`,
        """

        DeltaF_cov = (
            np.diag(self.F_cov)[:, None] + np.diag(self.F_cov)[None, :] - 2 * self.F_cov
        )
        return np.sqrt(DeltaF_cov)


def _dF_to_F(dF, num_conf):
    if dF.ndim == 1:
        F = jnp.concatenate([jnp.zeros((1,)), dF])
    elif dF.ndim == 2:
        F = jnp.concatenate([jnp.zeros((dF.shape[0], 1)), dF], axis=1)
    pi = num_conf / num_conf.sum()
    F = F - jnp.sum(pi * F, axis=-1, keepdims=True)
    return F


def _compute_loss_joint_likelihood_of_dF(dF, energy, num_conf, mean_prior, prec_prior):
    """
    Compute the loss function of dF based on the joint likelihood.

    The logarithm of the joint likelihood of dF scales with the number of configurations and the number of states. To make the loss function semi-invariant to the number of configurations and the number of states, we divide the log likelihood by the total number of configurations and the number of states. This helps to set a single tolerance for the optimization algorithm used to compute the MAP estimate.

    See the doc of _compute_log_joint_likelihood_of_dF for more details on the arguments and the return value.
    """

    loss = -_compute_log_joint_likelihood_of_dF(
        dF, energy, num_conf, mean_prior, prec_prior
    )
    loss = loss / (energy.shape[1] + energy.shape[0] - 1)
    return loss


def _compute_log_joint_likelihood_of_dF(dF, energy, num_conf, mean_prior, prec_prior):
    """
    Compute the logarithm of the joint likelihood of dF when the prior is a normal distribution.

    The joint likelihood is defined by the right hand side of Eq. (9) in the reference paper.

    Arguments:
        dF (jnp.ndarray): Free energy differences
        energy (jnp.ndarray): Energy matrix
        num_conf (jnp.ndarray): Number of configurations in each state
        mean_prior (jnp.ndarray): Mean of the prior
        prec_prior (jnp.ndarray): Precision matrix of the prior

    Returns:
        jnp.ndarray: Logarithm of the joint likelihood of dF

    """

    logp = -0.5 * jnp.dot(dF - mean_prior, jnp.dot(prec_prior, dF - mean_prior))
    logp = logp + _compute_log_likelihood_of_dF(dF, energy, num_conf)
    return logp


@partial(value_and_grad, argnums=1)
def _compute_elbo_loss(rng_key, raw_params, mean, kernel, data):
    energy = data["energy"]
    num_conf = data["num_conf"]
    state_cv = data["state_cv"]
    dF_prec_ll = data["dF_prec_ll"]
    dF_mean_ll = data["dF_mean_ll"]

    params = _params_from_raw(raw_params)
    mean_prior = mean(params["mean"], state_cv)
    cov_prior = kernel(params["kernel"], state_cv)
    mu_prop, cov_prop = _compute_proposal_dist(
        mean_prior, cov_prior, dF_mean_ll, dF_prec_ll
    )
    dFs = random.multivariate_normal(rng_key, mu_prop, cov_prop, shape=(1024,))
    Fs = jnp.concatenate([jnp.zeros((dFs.shape[0], 1)), dFs], axis=1)
    elbo = jax.vmap(_compute_log_likelihood_of_F, in_axes=(0, None, None))(
        Fs, energy, num_conf
    )
    elbo = jnp.mean(elbo)
    elbo = elbo - _compute_kl_divergence(mu_prop, cov_prop, mean_prior, cov_prior)
    return -elbo


def _compute_kl_divergence(mu0, cov0, mu1, cov1):
    L0 = jnp.linalg.cholesky(cov0)
    L1 = jnp.linalg.cholesky(cov1)

    M = jax.scipy.linalg.solve_triangular(L1, L0, lower=True)
    y = jax.scipy.linalg.solve_triangular(L1, mu1 - mu0, lower=True)

    kl = 0.5 * (
        jnp.sum(M**2)
        + jnp.sum(y**2)
        - mu0.shape[0]
        + 2 * jnp.sum(jnp.log(jnp.diag(L1)) - jnp.log(jnp.diag(L0)))
    )
    return kl


def _compute_proposal_dist(mean_prior, cov_prior, dF_mean_ll, dF_prec_ll):
    prec_prior = jnp.linalg.inv(cov_prior)
    prec = dF_prec_ll + prec_prior
    cov = jnp.linalg.inv(prec)
    mu = jnp.dot(cov, jnp.dot(dF_prec_ll, dF_mean_ll) + jnp.dot(prec_prior, mean_prior))
    return mu, cov


def _print_params(params):
    res = "beta: "
    for i in range(params["mean"]["beta"].shape[0]):
        res += f'{params["mean"]["beta"][i].item():.4f}, '

    res += f'scale: {params["kernel"]["scale"].item():.4f}, '

    res += "l_scale: "
    for i in range(params["kernel"]["length_scale"].shape[0]):
        res += f'{params["kernel"]["length_scale"][i].item():.4f}, '

    if "alpha" in params["kernel"].keys():
        res += f'alpha: {params["kernel"]["alpha"].item():.4f}, '

    res += "dscale: "
    for i in range(params["kernel"]["dscale"].shape[0]):
        res += f'{params["kernel"]["dscale"][i].item():.4f}, '
    return res


def _expand(x, order):
    xx = [jnp.ones((x.shape[0], 1))]
    for i in range(order):
        xx.append(x ** (i + 1))
    xx = jnp.concatenate(xx, axis=-1)
    return xx


def _mean(params, x, order):
    xx = _expand(x, order)
    return jnp.sum(params["beta"] * xx, axis=-1)


def _params_from_raw(raw_params):
    params = {}
    params["mean"] = raw_params["mean"]
    params["kernel"] = _kernel_params_from_raw(raw_params["kernel"])
    return params


def _params_to_raw(params):
    raw_params = {}
    raw_params["mean"] = params["mean"]
    raw_params["kernel"] = _kernel_params_to_raw(params["kernel"])
    return raw_params


def _init_mean_params(order, dF, state_cv):
    x = _expand(state_cv, order)
    beta = jnp.linalg.lstsq(x, dF, rcond=None)[0]
    params = {"beta": beta}
    return params


def _init_params(mean_order, kernel_name, dF, state_cv, num_conf):
    params = {}
    params["mean"] = _init_mean_params(mean_order, dF, state_cv)
    params["kernel"] = _init_kernel_params(kernel_name, dF, state_cv, num_conf)
    return params


def _init_kernel_params(kernel_name, dF, state_cv, num_conf):
    params = {}
    params["scale"] = jnp.std(dF)
    params["length_scale"] = (state_cv.max(0) - state_cv.min(0)) / state_cv.shape[0]
    params["dscale"] = jnp.ones_like(dF) * jnp.std(dF)

    if kernel_name == "RQ":
        params["alpha"] = jnp.ones((1,)) * 10

    return params


def _kernel_RQ(params, x):
    scale = params["scale"]
    length_scale = params["length_scale"]
    dscale = params["dscale"]
    alpha = params["alpha"]
    x = x / length_scale
    ds = _compute_squared_distance(x)
    return scale**2 * (1 + ds / (2 * alpha)) ** (-alpha) + dscale**2 * jnp.eye(
        ds.shape[0]
    )


def _kernel_SE(params, x):
    scale = params["scale"]
    length_scale = params["length_scale"]
    dscale = params["dscale"]
    x = x / length_scale
    ds = _compute_squared_distance(x)
    return scale**2 * jnp.exp(-0.5 * ds) + dscale**2 * jnp.eye(ds.shape[0])


def _kernel_Matern52(params, x):
    scale = params["scale"]
    length_scale = params["length_scale"]
    dscale = params["dscale"]
    x = x / length_scale
    ds = _compute_squared_distance(x)
    d = jnp.sqrt(ds + 1e-18)
    return scale**2 * (1 + jnp.sqrt(5.0) * d + 5.0 / 3.0 * ds) * jnp.exp(
        -jnp.sqrt(5.0) * d
    ) + dscale**2 * jnp.eye(ds.shape[0])


def _kernel_Matern32(params, x):
    scale = params["scale"]
    length_scale = params["length_scale"]
    dscale = params["dscale"]
    x = x / length_scale
    ds = _compute_squared_distance(x)
    d = jnp.sqrt(ds + 1e-18)
    return scale**2 * (1 + jnp.sqrt(3.0) * d) * jnp.exp(
        -jnp.sqrt(3.0) * d
    ) + dscale**2 * jnp.eye(ds.shape[0])


def _kernel_params_from_raw(raw_params):
    params = {}
    params["scale"] = jax.nn.softplus(raw_params["raw_scale"])
    params["length_scale"] = jax.nn.softplus(raw_params["raw_length_scale"])
    params["dscale"] = jax.nn.softplus(raw_params["raw_dscale"])
    if "raw_alpha" in raw_params.keys():
        params["alpha"] = jax.nn.softplus(raw_params["raw_alpha"])
    return params


def _kernel_params_to_raw(params):
    raw_params = {}
    raw_params["raw_scale"] = jnp.log(jnp.exp(params["scale"]) - 1)
    raw_params["raw_length_scale"] = jnp.log(jnp.exp(params["length_scale"]) - 1)
    raw_params["raw_dscale"] = jnp.log(jnp.exp(params["dscale"]) - 1)
    if "alpha" in params.keys():
        raw_params["raw_alpha"] = jnp.log(jnp.exp(params["alpha"]) - 1)
    return raw_params


def _compute_squared_distance(x):
    x1 = x[:, None, :]
    x2 = x[None, :, :]
    return jnp.sum((x1 - x2) ** 2, axis=-1)


def _sample_from_logdensity(
    rng_key, init_dF, logdensity, warmup_steps, num_samples, verbose
):
    ## warmup to find step size and mass matrix
    warmup = blackjax.window_adaptation(
        blackjax.nuts,
        logdensity,
        is_mass_matrix_diagonal=False,
        progress_bar=verbose,
    )
    rng_key, subkey = random.split(rng_key)
    (state, parameters), _ = warmup.run(subkey, init_dF, num_steps=warmup_steps)

    print("Sample using the NUTS sampler")

    ## sample using nuts

    # ## Use the blackjax.util.run_inference_algorithm function to run the nuts algorithm
    # ## so that we can have a progress bar. It is a wrap of _sample_loop.
    # alg = blackjax.nuts(logdensity, **parameters)
    # _, states, _ = blackjax.util.run_inference_algorithm(
    #     rng_key, state, alg, num_samples, progress_bar=verbose
    # )

    ## sample using nuts
    rng_key, subkey = random.split(rng_key)
    kernel = blackjax.nuts(logdensity, **parameters).step
    states = _sample_loop(subkey, kernel, state, num_samples)

    return states.position


def _sample_loop(rng_key, kernel, init_state, num_samples):
    @jax.jit
    def one_step(state, rng_key):
        state, _ = kernel(rng_key, state)
        return state, state

    keys = jax.random.split(rng_key, num_samples)
    _, states = jax.lax.scan(one_step, init_state, keys)
    return states



import pytest
from pytest import approx
from bayesmbar import BayesBAR

@pytest.mark.parametrize("method", ["Newton", "L-BFGS-B"])
def test_BayesBAR(setup_bar_data, method):
    energy, num_conf, DeltaF_ref = setup_bar_data
    energy = energy[0:2]
    num_conf = num_conf[0:2]
    bar = BayesBAR(
        energy,
        num_conf,
        method=method,        
        verbose=False,        
    )
    assert bar.DeltaF_mode.item() == approx(DeltaF_ref, abs = 1e-1)
    assert bar.DeltaF_mean.item() == approx(DeltaF_ref, abs = 1e-1)



import pytest
from pytest import approx
from bayesmbar import BayesMBAR

@pytest.mark.parametrize("method", ["Newton", "L-BFGS-B"])
def test_BayesMBAR(setup_mbar_data, method):
    energy, num_conf, F_ref, energy_p, F_ref_p = setup_mbar_data
    mbar = BayesMBAR(
        energy,
        num_conf,
        verbose=True,
        method=method,
    )
    assert mbar.F_mode == approx(F_ref, abs = 1e-1)
    assert mbar.F_mean == approx(F_ref, abs = 1e-1)

    #results = fastmbar.calculate_free_energies_of_perturbed_states(energy_p)
    #results['F'] = results['F'] - results['F'].mean()
    #assert results['F'] == approx(F_ref_p, abs = 1e-1)

# import pytest
# from pytest import approx
# import jax.numpy as jnp
# from jax import random
# from bayesmbar import BayesMBAR


# # @pytest.fixture
# def setup_data():
#     num_states = 5
#     key = random.PRNGKey(0)
#     key, subkey = random.split(key)
#     num_conf = random.randint(subkey, (num_states,), 50, 100)

#     key, subkey = random.split(key)
#     mu = random.uniform(subkey, (num_states,), jnp.float64, 0, 2)

#     key, subkey = random.split(key)
#     sigma = random.uniform(subkey, (num_states,), jnp.float64, 1, 3)

#     ## draw samples from each state and
#     ## calculate energies of each sample in all states
#     Xs = []
#     for i in range(num_states):
#         key, subkey = random.split(key)
#         Xs.append(random.normal(subkey, (num_conf[i],)) * sigma[i] + mu[i])

#     Xs = jnp.concatenate(Xs)
#     Xs = Xs.reshape((-1, 1))
#     energy = 0.5 * ((Xs - mu) / sigma) ** 2
#     energy = energy.T

#     F_ref = -jnp.log(sigma)
#     pi = num_conf / num_conf.sum()
#     F_ref = F_ref - jnp.sum(pi * F_ref)

#     return energy, num_conf, F_ref, key


# energy, num_conf, F_ref, key = setup_data()

# num_states = len(num_conf)
# key, subkey = random.split(key)
# cv = random.normal(subkey, (num_states, 1))
# d2 = bayesmbar._compute_squared_distance(cv)
# params = {'scale': 1.0, 'length_scale': 1.0}

# key, subkey = random.split(key)
# mbar = BayesMBAR(
#     energy, num_conf, subkey, 'normal', cv, 'SE', sample_size=2000, warmup_steps=200
# )

# @pytest.mark.parametrize("method", ["Newton", "L-BFGS-B"])
# @pytest.mark.parametrize("bootstrap", [False, True])
# def test_BayesMBAR(setup_data, method, bootstrap):
#     energy, num_conf, F_ref = setup_data
#     fastmbar = FastMBAR(
#         energy,
#         num_conf,
#         cuda=False,
#         cuda_batch_mode=False,
#         bootstrap=bootstrap,
#         verbose=False,
#         method=method,
#     )
#     # print(fastmbar.F)
#     # print(F_ref)
#     assert fastmbar.F == approx(F_ref, abs = 1e-1)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "BayesMBAR"
copyright = "2024, Xinqiang Ding"
author = "Xinqiang Ding"
# release = '0.0.1'

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath("../../"))
sys.path.insert(0, os.path.abspath("../../../"))


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "nbsphinx",
#    "autoapi.extension",
]


templates_path = ["_templates"]
exclude_patterns = []

# autodoc_default_options = {
#     'member-order': 'bysource',
# }
autoclass_content = 'both'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_logo = "_static/logo.png"
html_static_path = ["_static"]
html_theme_options = {
    'show_toc_level': 2,
    'repository_url': 'https://github.com/DingGroup/BayesMBAR',
    'use_repository_button': True,     # add a "link to repository" button
    'navigation_with_keys': False,
}

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


import pytest
from pytest import approx
from bayesmbar import FastMBAR


@pytest.mark.parametrize("method", ["Newton", "L-BFGS-B"])
@pytest.mark.parametrize("bootstrap", [False, True])
def test_FastMBAR(setup_mbar_data, method, bootstrap):
    energy, num_conf, F_ref, energy_p, F_ref_p = setup_mbar_data
    fastmbar = FastMBAR(
        energy,
        num_conf,
        bootstrap=bootstrap,
        bootstrap_num_rep = 30,
        verbose=False,
        method=method,
    )
    assert fastmbar.F == approx(F_ref, abs = 2e-1)

    results = fastmbar.calculate_free_energies_of_perturbed_states(energy_p)
    results['F'] = results['F'] - results['F'].mean()
    assert results['F'] == approx(F_ref_p, abs = 2e-1)


import math
import random
import numpy as np
from numpy import ndarray
import scipy.optimize as optimize
import scipy.integrate as integrate
import jax
from jax import hessian, jit, value_and_grad
import jax.numpy as jnp
from .utils import fmin_newton
from tqdm import tqdm

jax.config.update("jax_enable_x64", True)


class BayesBAR:
    """Bayesian Bennett acceptance ratio method"""

    def __init__(
        self,
        energy: ndarray,
        num_conf: ndarray,
        sample_size: int = 1000,
        method: str = "Newton",
        verbose=False,
    ):
        """
        Initialize the BayesBAR class.

        Parameters
        ----------
        energy : ndarray
            An energy matrix in reduced units. Its size should be 2xN, where N is the total number of samples from the two states.
        num_conf : ndarray
            Number of configurations in each state. Its size should be (2,).
        sample_size : int, optional
            The number of samples from the posterior distribution. Defaults to 1000.
        method : str, optional
            Optimization method for finding the mode. Options are "Newton" or "L-BFGS-B". Defaults to "Newton".
        verbose : bool, optional
            Whether to print running information. Defaults to False.
        """

        assert (
            energy.shape[0] == 2
        ), f"The energy matrix has {energy.shape[0]} rows. It has to have 2 rows."

        assert (
            len(num_conf) == 2
        ), f"The size of num_conf is {len(num_conf)}. It has to be 2."

        assert (
            energy.shape[1] == num_conf[0] + num_conf[1]
        ), f"The energy matrix has {energy.shape[1]} columns, but the sum of num_conf is {num_conf[0] + num_conf[1]}. The number of columns in the energy matrix must equal the sum of num_conf."

        self.energy = jnp.float64(energy)
        self.num_conf = jnp.int32(num_conf)

        self.n_0, self.n_1 = self.num_conf
        self.n = jnp.sum(self.num_conf)

        ## find the posterior mode which corresponds to the BAR solution
        dF_init = jnp.zeros(1, dtype=jnp.float64)
        dF_init = jnp.mean(self.energy[1] - self.energy[0]).reshape((-1,))

        if method == "Newton":
            f = jit(value_and_grad(_compute_loss))
            hess = jit(hessian(_compute_loss))
            res = fmin_newton(f, hess, dF_init, args=(self.energy, self.num_conf), verbose=verbose)
        elif method == "L-BFGS-B":
            options = {"disp": verbose, "gtol": 1e-8}
            f = jit(value_and_grad(_compute_loss))
            res = optimize.minimize(
                lambda x: [np.array(r) for r in f(x, self.energy, self.num_conf)],
                dF_init,
                jac=True,
                method="L-BFGS-B",
                tol=1e-12,
                options=options,
            )
        else:
            raise ValueError(
                f"Method {method} is not supported. It must be 'Newton' or 'L-BFGS-B'."
            )

        self.dF_mode = res["x"]

        ## compute posterior mean and standard deviation using numerical integration
        self.dF_mean, self.dF_std = _compute_posterior_mean_and_std(
            self.dF_mode, self.energy, self.num_conf
        )

        ## sampling from the posterior distribution
        self.sample_size = sample_size
        if self.sample_size > 0:
            self.dF_samples = _sample_from_posterior(
                self.dF_mode,
                self.dF_std,
                self.energy,
                self.num_conf,
                self.sample_size,
            )

        ## compute asymptotic standard deviation
        H = hessian(_compute_logp)(self.dF_mode, self.energy, self.num_conf)
        _dF_var_asymptotic = -1.0 / H - 1.0 / self.n_0 - 1.0 / self.n_1
        self._dF_std_asymptotic = jnp.reshape(jnp.sqrt(_dF_var_asymptotic), ())

        ## Bennett's uncertainty
        du = (
            self.energy[1, :]
            - self.energy[0, :]
            - jnp.log(self.n_1 / self.n_0)
            - self.dF_mode
        )
        f0 = jax.nn.sigmoid(-du[0 : self.n_0])
        f1 = jax.nn.sigmoid(du[self.n_0 :])
        _dF_var_bennett = (
            jnp.mean(f0**2) / (self.n_0 * jnp.mean(f0) ** 2)
            + jnp.mean(f1**2) / (self.n_1 * jnp.mean(f1) ** 2)
            - 1.0 / self.n_0
            - 1.0 / self.n_1
        )
        self._dF_std_bennett = jnp.sqrt(_dF_var_bennett)

    @property
    def DeltaF_mode(self) -> ndarray:
        """The posterior mode of the free energy difference."""
        return np.array(jax.device_put(self.dF_mode, jax.devices("cpu")[0]))

    @property
    def DeltaF_mean(self) -> ndarray:
        """The posterior mean of the free energy difference."""
        return np.array(jax.device_put(self.dF_mean, jax.devices("cpu")[0]))

    @property
    def DeltaF_std(self) -> ndarray:
        """The posterior standard deviation of the free energy difference."""
        return np.array(jax.device_put(self.dF_std, jax.devices("cpu")[0]))

    @property
    def DeltaF_samples(self) -> ndarray:
        """The samples from the posterior distribution of the free energy difference."""
        return np.array(jax.device_put(self.dF_samples, jax.devices("cpu")[0]))


@jit
def _compute_logp(dF, energy, num_conf):
    n_0, n_1 = num_conf
    du = energy[1, :] - energy[0, :] - jnp.log(n_1 / n_0)
    logp = n_1 * dF - jnp.logaddexp(jnp.zeros(1), dF - du).sum()
    return jnp.reshape(logp, ())


@jit
def _compute_loss(dF, energy, num_conf):
    logp = _compute_logp(dF, energy, num_conf)
    loss = -logp / num_conf.sum()
    return loss


@jit
def _compute_posterior(dF, energy, num_conf, dF_mode):
    logp_max = _compute_logp(dF_mode, energy, num_conf)

    ## The prior is chosen to be the uniform distribution over R,
    ## so the posterior is equal to the likelihood. The likelihood
    ## is shifted down by a constant which is the self.logp_max

    return jnp.exp(_compute_logp(dF, energy, num_conf) - logp_max)


def _compute_posterior_mean_and_std(dF_mode, energy, num_conf):
    ## compute the normalization constant Z
    def f(dF):
        return _compute_posterior(dF, energy, num_conf, dF_mode)

    Z, Z_err = integrate.quad(jit(f), -np.inf, np.inf)

    ## posterior mean
    def f(dF):
        return dF * _compute_posterior(dF, energy, num_conf, dF_mode)

    dF, dF_err = integrate.quad(jit(f), -np.inf, np.inf)
    dF_mean = dF / Z

    ## posterior standard deviation
    def f(dF):
        return (dF - dF_mean) ** 2 * _compute_posterior(dF, energy, num_conf, dF_mode)

    dF_var, dF_var_err = integrate.quad(jit(f), -np.inf, np.inf)
    dF_var = dF_var / Z
    dF_std = math.sqrt(dF_var)
    return dF_mean, dF_std


def _sample_from_posterior(dF_mode, dF_std, energy, num_conf, size):
    """sample from the posterior using slice sampling
    (https://www.jstor.org/stable/3448413 )
    """

    ## dF_std is used as the estimate of the typical size of a slice.
    width = dF_std

    ## the size of a slice will be limited to max_size*width
    max_size = 3

    ## start from the posterior mode
    x0 = dF_mode
    samples = [x0]

    for _ in tqdm(range(size - 1)):
        ## sample the auxiliarxy random variable
        logp = _compute_logp(x0, energy, num_conf)
        z = logp - random.expovariate(1.0)

        ## find the slice interval using the "stepping out" procedure
        u = random.uniform(0.0, 1.0)
        L = x0 - u * width
        R = L + width

        v = random.uniform(0.0, 1.0)
        J = math.floor(v * max_size)
        K = max_size - 1 - J

        while J > 0 and z < _compute_logp(L, energy, num_conf):
            L = L - width
            J = J - 1

        while K > 0 and z < _compute_logp(R, energy, num_conf):
            R = R + width
            K = K - 1

        ## sampling from the interval using the "shrinkage" procedure
        while True:
            x1 = random.uniform(L, R)
            if z < _compute_logp(x1, energy, num_conf):
                samples.append(x1)
                x0 = x1
                break
            if x1 < x0:
                L = x1
            else:
                R = x1

    return jnp.array(samples).reshape(-1)


import jax.random as jr
import jax.numpy as jnp
import numpy as np
import pytest

@pytest.fixture
def setup_mbar_data():
    num_states = 5
    key = jr.PRNGKey(0)
    key, subkey = jr.split(key)
    num_conf = jr.randint(subkey, (num_states,), 50, 100)

    key, subkey = jr.split(key)
    mu = jr.uniform(subkey, (num_states,), jnp.float64, 0, 2)

    key, subkey = jr.split(key)
    sigma = jr.uniform(subkey, (num_states,), jnp.float64, 1, 3)

    ## draw samples from each state and
    ## calculate energies of each sample in all states
    Xs = []
    for i in range(num_states):
        key, subkey = jr.split(key)
        Xs.append(jr.normal(subkey, (num_conf[i],)) * sigma[i] + mu[i])

    Xs = jnp.concatenate(Xs)
    Xs = Xs.reshape((-1, 1))
    energy = 0.5 * ((Xs - mu) / sigma) ** 2
    energy = energy.T

    F_ref = -jnp.log(sigma)
    pi = num_conf / num_conf.sum()
    F_ref = F_ref - jnp.sum(pi * F_ref)

    ## perturbed states
    key, subkey = jr.split(key)
    num_states_p = 8
    mu_p = jr.uniform(subkey, (num_states_p,), jnp.float64, 0, 2)
    sigma_p = jr.uniform(subkey, (num_states_p,), jnp.float64, 0, 4)

    energy_p = 0.5 * ((Xs - mu_p) / sigma_p) ** 2
    energy_p = energy_p.T

    F_ref_p = -jnp.log(sigma_p)
    F_ref_p = F_ref_p - jnp.mean(F_ref_p)

    return np.array(energy), np.array(num_conf), np.array(F_ref), np.array(energy_p), np.array(F_ref_p)


@pytest.fixture
def setup_bar_data():
    num_states = 2
    key = jr.PRNGKey(0)
    key, subkey = jr.split(key)
    num_conf = jr.randint(subkey, (num_states,), 50, 100)

    key, subkey = jr.split(key)
    mu = jr.uniform(subkey, (num_states,), jnp.float64, 0, 2)

    key, subkey = jr.split(key)
    sigma = jr.uniform(subkey, (num_states,), jnp.float64, 1, 3)

    ## draw samples from each state and
    ## calculate energies of each sample in all states
    Xs = []
    for i in range(num_states):
        key, subkey = jr.split(key)
        Xs.append(jr.normal(subkey, (num_conf[i],)) * sigma[i] + mu[i])

    Xs = jnp.concatenate(Xs)
    Xs = Xs.reshape((-1, 1))
    energy = 0.5 * ((Xs - mu) / sigma) ** 2
    energy = energy.T

    F_ref = -jnp.log(sigma)
    pi = num_conf / num_conf.sum()
    F_ref = F_ref - jnp.sum(pi * F_ref)

    DeltaF = F_ref[1] - F_ref[0]

    return np.array(energy), np.array(num_conf), np.array(DeltaF)


import pytest
from pytest import approx
from bayesmbar import CBayesMBAR
import numpy as np


@pytest.fixture
def setup_data_cbmbar():
    ka, kb, kc, kd = 4, 9, 16, 25
    ## equilibrium positions of three two-dimensional harmonic oscillators
    mu = {
        "a": np.array([-1.0, 1.0]),
        "b": np.array([1.0, 1.0]),
        "c": np.array([1.0, -1.0]),
        "d": np.array([-1.0, -1.0]),
    }

    ## spring constants
    k = {
        "a": np.ones(2) * ka,
        "b": np.ones(2) * kb,
        "c": np.ones(2) * kc,
        "d": np.ones(2) * kd,
    }

    ## interpolate the equilibrium positions and spring constants between pairs of states
    pairs = [("a", "b"), ("b", "c"), ("c", "d"), ("d", "a"), ("a", "c"), ("b", "d")]
    for s1, s2 in pairs:
        for idx in range(1, 4):
            mu[(s1, s2, idx)] = mu[s1] + (mu[s2] - mu[s1]) / 4 * idx
            k[(s1, s2, idx)] = k[s1] + (k[s2] - k[s1]) / 4 * idx

    n = 2000

    ## compute the free energy difference between pairs of states analytically
    def k_to_sigma(k):
        return np.sqrt(1.0 / k)

    deltaF_ref = {}
    for s1, s2 in pairs:
        deltaF_ref[(s1, s2)] = (
            -np.log(k_to_sigma(k[s2])).sum() + np.log(k_to_sigma(k[s1])).sum()
        )

    ## sample configurations from all states including both endpoints and intermediates
    np.random.seed(0)
    x = {}
    for s in mu.keys():
        x[s] = np.random.normal(mu[s], k_to_sigma(k[s]), (n, 2))

    u_list = []
    for s1, s2 in pairs:
        key = [s1] + [(s1, s2, idx) for idx in range(1, 4)] + [s2]
        xs = np.concatenate([x[s] for s in key])
        u = np.stack(
            [np.sum(0.5 * k[s] * (xs - mu[s]) ** 2, axis=1) for s in key], axis=0
        )
        u_list.append(u)

    num_conf_list = [
        [u.shape[1] // u.shape[0] for i in range(u.shape[0])] for u in u_list
    ]
    identical_states = (
        [(0, 0), (3, 4), (4, 0)],
        [(0, 4), (1, 0), (5, 0)],
        [(1, 4), (2, 0), (4, 4)],
        [(2, 4), (3, 0), (5, 4)],
    )

    return u_list, num_conf_list, identical_states, pairs, deltaF_ref


@pytest.mark.parametrize("method", ["Newton", "L-BFGS-B"])
def test_CBayesMBAR(setup_data_cbmbar, method):
    u_list, num_conf_list, identical_states, pairs, deltaF_ref = setup_data_cbmbar
    cbmbar = CBayesMBAR(
        u_list, num_conf_list, identical_states, method=method, random_seed=0
    )

    deltaF_cbmbar = {}
    for i in range(len(pairs)):
        s1, s2 = pairs[i]
        deltaF_cbmbar[(s1, s2)] = cbmbar.DeltaF_mode[i][0, -1].item()

    assert deltaF_cbmbar == approx(deltaF_ref, abs=1e-1)
